/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2019  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QGridLayout>
#include <QLineEdit>
#include <QWidget>
#include <QSpacerItem>
#include <assert.h>

#include "dataitemwidget.h"

#include "dataeditorwidget.h"

// necessary to control widths if all columns are to line up.
// something that doesn't use a hard constant is preferable
static const int						EDITOR_ITEM_WIDTH	= 20;
static const int						EDITOR_TITLE_WIDTH	= 100;
static const int						EDITOR_EDIT_WIDTH	= 100;

TDataEditorWidget::TDataEditorWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	
	widget_layout = new QGridLayout(this);
	widget_layout->setContentsMargins(0,0,0,0);

	d_items_widget = new QWidget(this);
	widget_layout->addWidget(d_items_widget,1,0,1,2);
	
	d_items_widget_vlayout = new QVBoxLayout(d_items_widget);
	d_items_widget_vlayout->setContentsMargins(0,0,0,0);
	d_items_widget_vlayout->setSpacing(0);
	
	d_items_widget_vspacer = new QSpacerItem(0, 0, QSizePolicy::Minimum, QSizePolicy::Expanding);
	d_items_widget_vlayout->addItem(d_items_widget_vspacer);

	d_active_line_edit = new QLineEdit(this);
	widget_layout->addWidget(d_active_line_edit,0,0,1,2);
	
	// defaults
	d_scale_resolution = 1000;
	
	connect(d_active_line_edit,SIGNAL(textEdited(const QString&)),this,SLOT(Simple_Editor_Changed(void)));

}

TDataEditorWidget::~TDataEditorWidget(void)
{
}

QString TDataEditorWidget::Simple_Editor_Data(void) const
{
	return d_active_line_edit->text();
}

QString TDataEditorWidget::Item_Widget_Data(
	const int							index) const
{
	assert(static_cast<unsigned int>(index) < d_data_widget_items.size());
	
	return d_data_widget_items[index]->Hex_Data();
}

void TDataEditorWidget::Reset(
	const bool							use_simple_editor)
{
	std::vector<TDataItemWidget*>::iterator iter;

	d_active_items = 0;
	d_data_item_changed = false;
	
	d_simple_editor = use_simple_editor;
	
	for(iter = d_data_widget_items.begin();iter != d_data_widget_items.end();++iter)
	{
		(*iter)->setVisible(false);
	}
	
	if(d_simple_editor)
	{
		d_active_line_edit->setEnabled(true);
	}
	else
	{
		d_active_line_edit->setEnabled(false);
	}
}

void TDataEditorWidget::Set_Item_Widget_Count(
	const int							count)
{
	TDataItemWidget						*item;
	int									cntr;
	std::vector<TDataItemWidget*>::iterator iter;
	bool								restore_vspacer(false);

	assert(!(count < 0));
	
	if(!(count < static_cast<int>(d_data_widget_items.size())))
	{
		d_items_widget_vlayout->removeItem(d_items_widget_vspacer);
		restore_vspacer = true;
	}
	
	while(count > static_cast<int>(d_data_widget_items.size()))
	{
		item = new TDataItemWidget(d_data_widget_items.size(),d_items_widget);
		
		item->Set_Item_Width(EDITOR_ITEM_WIDTH);
		item->Set_Title_Width(EDITOR_TITLE_WIDTH);
		item->Set_Edit_Width(EDITOR_EDIT_WIDTH);
		
		d_items_widget_vlayout->addWidget(item);
		
		connect(item,SIGNAL(Int_Edited(int,const QString&)),this,SLOT(Int_Edited(int,const QString&)));
		connect(item,SIGNAL(Hex_Edited(int,const QString&)),this,SLOT(Hex_Edited(int,const QString&)));
		connect(item,SIGNAL(Dec_Edited(int,const QString&)),this,SLOT(Dec_Edited(int,const QString&)));
		
		d_data_widget_items.push_back(item);
	}
	
	for(iter = d_data_widget_items.begin(),cntr = 0;iter != d_data_widget_items.end();++iter,++cntr)
	{
		(*iter)->setVisible(cntr < count);
	}
	
	if(restore_vspacer)
	{
		d_items_widget_vlayout->addItem(d_items_widget_vspacer);
	}
}
void TDataEditorWidget::Set_Item_Widget_Data(
	const int							index,
	const QChar							item,
	const int							value,
	const QString						&description,
	const bool							enable_edit,
	const bool							enable_float_value)
{
	TDataItemWidget						*data_item;
	double								dval(0.0);

	assert(static_cast<unsigned int>(index) < d_data_widget_items.size());
	
	data_item = d_data_widget_items[index];
	
	if(enable_float_value)
	{
		dval = static_cast<double>(value) / static_cast<double>(d_scale_resolution);
	}
	
	if(enable_edit)
	{
		data_item->setEnabled(true);
	}
	else
	{
		data_item->setEnabled(false);
	}
	
	data_item->Set_Item(item);
	data_item->Set_Integer(value);
	data_item->Set_Hexadecimal(value);
	data_item->Set_Value(dval);
	data_item->Set_Description(description);
	data_item->Enable_Value(enable_float_value);
}

void TDataEditorWidget::Set_Simple_Editor_Data(
	const QString						&text)
{
	d_active_line_edit->setText(text);
}

void TDataEditorWidget::Simple_Editor_Changed(void)
{
	d_data_item_changed = true;
}

void TDataEditorWidget::Int_Edited(
	int									id,
	const QString						&text)
{
	TDataItemWidget						*data_item;
	int									ival(0);
	double								dval(0.0);

	assert(static_cast<unsigned int>(id) < d_data_widget_items.size());
	
	data_item = d_data_widget_items[id];
	
	ival = text.toInt();
	
	if(data_item->Value_Enabled())
	{
		dval = static_cast<double>(ival) / static_cast<double>(d_scale_resolution);
		data_item->Set_Value(dval);
	}
	
	data_item->Set_Hexadecimal(ival);
	
	d_data_item_changed = true;
}

void TDataEditorWidget::Hex_Edited(
	int									id,
	const QString						&text)
{
	TDataItemWidget						*data_item;
	int									ival(0);
	double								dval(0.0);
	bool								valid;
	
	assert(static_cast<unsigned int>(id) < d_data_widget_items.size());
	
	data_item = d_data_widget_items[id];
	
	ival = static_cast<int>(text.toUInt(&valid,16));
	
	if(!valid)
	{
		return;
	}
	
	if(data_item->Value_Enabled())
	{
		dval = static_cast<double>(ival) / static_cast<double>(d_scale_resolution);
		data_item->Set_Value(dval);
	}
	
	data_item->Set_Integer(ival);
	
	d_data_item_changed = true;
}

void TDataEditorWidget::Dec_Edited(
	int									id,
	const QString						&text)
{
	TDataItemWidget						*data_item;
	int									ival(0);
	double								dval(0.0);
	
	assert(static_cast<unsigned int>(id) < d_data_widget_items.size());
	
	data_item = d_data_widget_items[id];
	
	dval = text.toDouble();
	
	if(dval < 0)
	{
		ival = static_cast<int>(dval * static_cast<double>(d_scale_resolution) - 0.5);
	}
	else
	{
		ival = static_cast<int>(dval * static_cast<double>(d_scale_resolution) + 0.5);
	}
	
	data_item->Set_Integer(ival);
	data_item->Set_Hexadecimal(ival);
	
	d_data_item_changed = true;
}
