/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2019  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QFrame>
#include <QGridLayout>
#include <QLabel>
#include <QLineEdit>
#include <QEvent>
#include <QKeyEvent>
#include <QSpacerItem>
#include <QSizePolicy>

#include "dataitemwidget.h"

TDataItemWidget::TDataItemWidget(
	const int							id,
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QFrame								*left_vline;
	QFrame								*lower_hline;
	QFrame								*right_vline;
	QFrame								*upper_hline;
	QGridLayout							*widget_layout;
	QSizePolicy							help_label_size_policy(QSizePolicy::Expanding,QSizePolicy::Preferred);

	d_id = id;
	
	help_label_size_policy.setHorizontalStretch(0);
	help_label_size_policy.setVerticalStretch(0);

	widget_layout = new QGridLayout(this);
	widget_layout->setVerticalSpacing(2);

	lower_hline = new QFrame(this);
	lower_hline->setFrameShape(QFrame::HLine);
	lower_hline->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(lower_hline,4,0,1,6);

	d_int_label = new QLabel(this);
	d_int_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	widget_layout->addWidget(d_int_label,1,2,1,1);

	d_int_value_edit = new QLineEdit(this);
	d_int_value_edit->setAlignment(Qt::AlignCenter);
	widget_layout->addWidget(d_int_value_edit,1,3,1,1);

	d_value_label = new QLabel(this);
	d_value_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	widget_layout->addWidget(d_value_label,3,2,1,1);

	left_vline = new QFrame(this);
	left_vline->setFrameShape(QFrame::VLine);
	left_vline->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(left_vline,1,1,3,1);

	upper_hline = new QFrame(this);
	upper_hline->setFrameShape(QFrame::HLine);
	upper_hline->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(upper_hline,0,0,1,6);

	d_help_label = new QLabel(this);
	d_help_label->setSizePolicy(help_label_size_policy);
	d_help_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	d_help_label->setWordWrap(true);
	widget_layout->addWidget(d_help_label,1,5,3,1);

	d_hex_value_edit = new QLineEdit(this);
	d_hex_value_edit->setAlignment(Qt::AlignCenter);
	widget_layout->addWidget(d_hex_value_edit,2,3,1,1);

	d_item_label = new QLabel(this);
	d_item_label->setMinimumSize(50,0);
	d_item_label->setAlignment(Qt::AlignCenter | Qt::AlignVCenter);//Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	widget_layout->addWidget(d_item_label,1,0,3,1);

	d_double_value_edit = new QLineEdit(this);
	d_double_value_edit->setAlignment(Qt::AlignCenter);
	widget_layout->addWidget(d_double_value_edit,3,3,1,1);

	d_hex_label = new QLabel(this);
	d_hex_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	widget_layout->addWidget(d_hex_label,2,2,1,1);

	right_vline = new QFrame(this);
	right_vline->setFrameShape(QFrame::VLine);
	right_vline->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(right_vline,1,4,3,1);

	d_int_label->setText(QStringLiteral("Integer:"));
	d_value_label->setText(QStringLiteral("Value:"));
	d_help_label->setText(QStringLiteral("Help Text"));
	d_item_label->setText(QStringLiteral("Item"));
	d_hex_label->setText(QStringLiteral("Hexadecimal:"));
	
	// event filters
//	this->Install_Event_Filters(event_parent);
	
	connect(d_int_value_edit,SIGNAL(textEdited(const QString&)),this,SLOT(Int_Edited(const QString&)));
	connect(d_hex_value_edit,SIGNAL(textEdited(const QString&)),this,SLOT(Hex_Edited(const QString&)));
	connect(d_double_value_edit,SIGNAL(textEdited(const QString&)),this,SLOT(Dec_Edited(const QString&)));
}

TDataItemWidget::~TDataItemWidget(void)
{
}

bool TDataItemWidget::Value_Enabled(void) const
{
	return d_double_value_edit->isEnabled();
}

QString TDataItemWidget::Hex_Data(void) const
{
	QString								data;
	QRegExp								entry_expression(QString("([0-9ABCDEF]{0,8})"));	// 0 to FFFFFFFF
	int									entry_index;

	// make sure data is valid
	data = d_hex_value_edit->text().toUpper().simplified();
	data.remove(' ');
	
	entry_index = entry_expression.indexIn(data);
	
	if(!(entry_index < 0) && entry_expression.captureCount() == 1)
	{
		data = entry_expression.cap(1);
	}
	else
	{
		data = QStringLiteral("0");
	}
	
	return data;
}

void TDataItemWidget::Set_Item_Width(
	const int							width)
{
	d_item_label->setFixedWidth(width);
}

void TDataItemWidget::Set_Title_Width(
	const int							width)
{
	d_int_label->setFixedWidth(width);
	d_hex_label->setFixedWidth(width);
	d_value_label->setFixedWidth(width);
}

void TDataItemWidget::Set_Edit_Width(
	const int							width)
{
	d_int_value_edit->setFixedWidth(width);
	d_hex_value_edit->setFixedWidth(width);
	d_double_value_edit->setFixedWidth(width);
}

void TDataItemWidget::Enable_Value(
	const bool							state)
{
	d_value_label->setEnabled(state);
	d_double_value_edit->setEnabled(state);
	
	if(!state)
	{
		d_double_value_edit->setText(QString());
	}
}

void TDataItemWidget::Set_Integer(
	const int							ival)
{
	d_int_value_edit->setText(QString("%1").arg(ival,0,10));
}

void TDataItemWidget::Set_Hexadecimal(
	const int							ival)
{
	QString								hex_text;
	int									length;
	
	hex_text = QString("%1").arg(ival,0,16);
	length = hex_text.length();
	
	if(length > 8)
	{
		hex_text = hex_text.mid(length - 8,8);
	}
	
	d_hex_value_edit->setText(hex_text.toUpper());
}

void TDataItemWidget::Set_Value(
	const double						&dval)
{
	d_double_value_edit->setText(QString("%1").arg(dval,0,'f',4));
}

void TDataItemWidget::Set_Item(
	const QChar							item)
{
	d_item_label->setText(QString("%1").arg(item));
}

void TDataItemWidget::Set_Description(
	const QString						&description)
{
	d_help_label->setText(description);
}

void TDataItemWidget::Int_Edited(
	const QString						&text)
{
	emit Int_Edited(d_id,text);
}

void TDataItemWidget::Hex_Edited(
	const QString						&text)
{
	emit Hex_Edited(d_id,text);
}

void TDataItemWidget::Dec_Edited(
	const QString						&text)
{
	emit Dec_Edited(d_id,text);
}
