/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2019  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QFrame>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QSizePolicy>
#include <QSpacerItem>
#include <QToolButton>
#include <QWidget>
#include <QToolBar>
#include <QComboBox>
#include <QAction>
#include <QSettings>
#include <QRegExp>
#include <QKeyEvent>
#include <QCloseEvent>
#include <QFileDialog>
#include <QDir>
#include <QDragEnterEvent>
#include <QDropEvent>
#include <QMimeData>
#include <QUrl>
#include <QList>
#include <QFont>
#include <QMessageBox>
#include <assert.h>

#include "../../core/messagebox.h"
#include "dataeditorwidget.h"
#include "filelinewidget.h"

#include "downloadeditor.h"

TDownloadEditor::TDownloadEditor(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QDialog(const_cast<QWidget*>(parent),flags)
{
	QToolBar							*toolbar;
	QWidget								*toolbar_spacer_widget;
	QFrame								*separator_hline;
	QVBoxLayout							*dialog_vlayout;
	QGridLayout							*dialog_layout;
	QHBoxLayout							*button_hlayout;
	QHBoxLayout							*file_group_hlayout;
	QPushButton							*close_button;
	QSizePolicy							file_name_size_policy(QSizePolicy::Expanding,QSizePolicy::Preferred);
	QSizePolicy							file_line_size_policy(QSizePolicy::Expanding,QSizePolicy::Expanding);
	QSizePolicy							data_editor_size_policy(QSizePolicy::Expanding,QSizePolicy::Expanding);
	QSpacerItem							*button_hspacer;
	QLabel								*copyright_label;
	QLabel								*theme_label;
	QFont								copyright_font;
	QDir								settings_path;
	int									margin_left,margin_top,margin_right,margin_bottom;
	
	d_application_theme.Set_Theme(TApplicationTheme::DEFAULT);
	
	this->setAcceptDrops(true);
	this->resize(1001,701);
	
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);
	
	file_line_size_policy.setHorizontalStretch(0);
	file_line_size_policy.setVerticalStretch(0);

	data_editor_size_policy.setHorizontalStretch(0);
	data_editor_size_policy.setVerticalStretch(0);
	
#ifdef Q_OS_WIN
	
	settings_path.setPath(qApp->applicationDirPath());
	
#else
	
	settings_path = QDir::home();
	
	if(!settings_path.exists(QStringLiteral(".downloadeditor")))
	{
		settings_path.mkdir(QStringLiteral(".downloadeditor"));
	}
	
	settings_path.cd(QStringLiteral(".downloadeditor"));
	
#endif
	
	d_settings = new QSettings(settings_path.absolutePath() + "/settings.ini",QSettings::IniFormat);

	dialog_vlayout = new QVBoxLayout(this);
	
	dialog_vlayout->getContentsMargins(&margin_left,&margin_top,&margin_right,&margin_bottom);
	dialog_vlayout->setContentsMargins(0,0,0,0);
	
	toolbar = new QToolBar(this);
	toolbar->setIconSize(QSize(32,32));
	
	d_load_file_action = new QAction(this);
	d_load_file_action->setToolTip(QStringLiteral("Load download file"));
	d_load_file_action->setShortcut(QStringLiteral("Ctrl+O"));
	d_load_file_action->setIcon(QIcon(":/toolbar/file_open.png"));
	
	d_save_file_action = new QAction(this);
	d_save_file_action->setToolTip(QStringLiteral("Save download file"));
	d_save_file_action->setShortcut(QStringLiteral("Ctrl+S"));
	d_save_file_action->setIcon(QIcon(":/toolbar/file_save.png"));

	toolbar->addAction(d_load_file_action);
	toolbar->addAction(d_save_file_action);
	
	toolbar_spacer_widget = new QWidget(this);
	toolbar_spacer_widget->setSizePolicy(file_name_size_policy);
	toolbar->addWidget(toolbar_spacer_widget);
	
	theme_label = new QLabel(this);
	toolbar->addWidget(theme_label);

	d_theme_combo = new QComboBox(this);
	d_theme_combo->setFocusPolicy(Qt::NoFocus);
	toolbar->addWidget(d_theme_combo);
	
	dialog_vlayout->addWidget(toolbar);

	dialog_layout = new QGridLayout;
	dialog_layout->setContentsMargins(margin_left,margin_top,margin_right,margin_bottom);

	file_group_hlayout = new QHBoxLayout;
	
	d_file_name_label = new QLabel(this);
	file_group_hlayout->addWidget(d_file_name_label);
	
	d_active_file_label = new QLabel(this);
	d_active_file_label->setSizePolicy(file_name_size_policy);
	file_group_hlayout->addWidget(d_active_file_label);

	dialog_layout->addLayout(file_group_hlayout,0,0,1,2);

	d_file_line_widget = new TFileLineWidget(this);
	d_file_line_widget->setMinimumWidth(100);
	d_file_line_widget->setMaximumWidth(320);
	d_file_line_widget->setSizePolicy(file_line_size_policy);
	dialog_layout->addWidget(d_file_line_widget,1,0,1,1);

	d_data_editor_widget = new TDataEditorWidget(this);
	d_data_editor_widget->setMinimumWidth(400);
	d_data_editor_widget->setSizePolicy(data_editor_size_policy);
	dialog_layout->addWidget(d_data_editor_widget,1,1,1,1);

	separator_hline = new QFrame(this);
	separator_hline->setFrameShape(QFrame::HLine);
	separator_hline->setFrameShadow(QFrame::Sunken);
	dialog_layout->addWidget(separator_hline,2,0,1,2);

	button_hlayout = new QHBoxLayout();
	
	copyright_label = new QLabel(this);
	copyright_label->setFont(copyright_font);
	button_hlayout->addWidget(copyright_label);

	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);

	close_button = new QPushButton(this);
	close_button->setAutoDefault(false);
	button_hlayout->addWidget(close_button);
	dialog_layout->addLayout(button_hlayout,3,0,1,2);
	
	dialog_vlayout->addLayout(dialog_layout);
	
	// defaults
	d_active_line_index = -1;
	d_data_changed = false;
	d_load_file_action->setEnabled(true);
	d_save_file_action->setEnabled(false);

	this->setWindowTitle(QStringLiteral("Sharpe32 Download Editor - Version 2.0"));
	
	d_theme_combo->addItems(d_application_theme.Theme_List());
	
	d_theme_combo->setCurrentIndex(d_settings->value(QStringLiteral("Theme_Index"),0).toInt());
	d_active_file_label->setText(d_settings->value(QStringLiteral("Last_File"),QString()).toString());
	d_file_line_widget->Set_Scale_Resolution(d_settings->value(QStringLiteral("Scale_Resolution"),10000).toInt());
	
	theme_label->setText(QStringLiteral("Theme:"));
	d_file_name_label->setText(QStringLiteral("Recent Download File:"));
	close_button->setText(QStringLiteral("Close"));
	copyright_label->setText(QStringLiteral("Copyright (C) 2019 Select Calibration Incorporated.  www.selectcalibration.ca"));
	
	if(d_application_theme.Is_Themes_Available())
	{
		this->Theme_Changed(d_theme_combo->currentIndex());
	}
	else
	{
		d_theme_combo->setCurrentIndex(0);
		d_theme_combo->setEnabled(false);
	}

	connect(d_load_file_action,SIGNAL(triggered(bool)),this,SLOT(Read_File(void)));
	connect(d_save_file_action,SIGNAL(triggered(bool)),this,SLOT(Write_File(void)));
	connect(d_theme_combo,SIGNAL(activated(int)),this,SLOT(Theme_Changed(int)));
	connect(close_button,SIGNAL(clicked(void)),this,SLOT(close(void)));

	connect(d_file_line_widget,SIGNAL(Line_Activated(int)),this,SLOT(Line_Activated(int)));
	connect(d_file_line_widget,SIGNAL(Scale_Resolution_Changed(int)),this,SLOT(Scale_Resolution_Changed(int)));
	
}

TDownloadEditor::~TDownloadEditor(void)
{
}

void TDownloadEditor::Read_File(
	const QString						&file_name)
{
	if(file_name.length())
	{
		d_active_line_index = -1;
		d_active_command_id = -1;
		
		if(d_file_line_widget->Read_File(file_name))
		{
			d_save_file_action->setEnabled(true);
		}
		
		d_active_file_label->setText(QDir::toNativeSeparators(file_name));
		d_file_name_label->setText(QStringLiteral("Active Download File:"));
		
		d_data_changed = false;
	}
}

void TDownloadEditor::Line_Activated(
	int									line_index)
{
	QString								line_text;
	
	// update existing editor data if it has changed
	if((d_active_line_index < 0) == false && d_data_editor_widget->Data_Item_Changed() == true)
	{
		this->Update_Active_Line();
		d_data_changed = true;
	}
	
	line_text = d_file_line_widget->Line(line_index);	
	this->Process_Line(line_text);
	
	d_active_line_index = line_index;
}

void TDownloadEditor::Theme_Changed(
	int									index)
{	
	Q_UNUSED(index);
	
	d_application_theme.Set_Theme(d_theme_combo->currentText());
}

void TDownloadEditor::closeEvent(
	QCloseEvent							*event)
{
	d_settings->setValue(QStringLiteral("Theme_Index"), d_theme_combo->currentIndex());
	d_settings->setValue(QStringLiteral("Last_File"), d_active_file_label->text());
	d_settings->setValue(QStringLiteral("Scale_Resolution"), d_file_line_widget->Scale_Resolution());

	if(d_data_changed)
	{
		TMessageBox						msgbox(this);
		int								result;
		
		msgbox.setText("Data Not Saved");
		msgbox.setInformativeText("The download file has been modified.  Do you want to save the changes?");
		msgbox.setDetailedText(QString());
		msgbox.setIcon(QMessageBox::Warning);
		msgbox.setStandardButtons(QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
		msgbox.setDefaultButton(QMessageBox::Yes);
		
		result = msgbox.exec();
		
		switch(result)
		{
			case QMessageBox::Yes:
				this->Write_File();
				break;
				
			case QMessageBox::No:
				break;
				
			case QMessageBox::Cancel:
			default:
				event->ignore();
				return;
		}
	}
	
	event->accept();
}

void TDownloadEditor::dragEnterEvent(
	QDragEnterEvent 					*drag_enter_event)
{
	if(drag_enter_event->mimeData()->hasFormat("text/uri-list"))
	{
		drag_enter_event->acceptProposedAction();
	}
}

void TDownloadEditor::dropEvent(
	QDropEvent 							*drop_event)
{
	QList<QUrl>							urls;
	QString								file_name;
	
	urls = drop_event->mimeData()->urls();
	
	if(urls.size() == 1)
	{
		file_name = urls[0].toLocalFile();
		
		d_active_line_index = -1;
		d_active_command_id = -1;
		
		if(d_file_line_widget->Read_File(file_name))
		{
			d_save_file_action->setEnabled(true);
		}
		
		d_active_file_label->setText(QDir::toNativeSeparators(file_name));
		d_file_name_label->setText(QStringLiteral("Active Download File:"));

	}
}

void TDownloadEditor::Key_Pressed(
	int									key)
{
	if(key == Qt::Key_Up)
	{
		d_file_line_widget->Decrement_Active_Line();
	}
	else if(key == Qt::Key_Down)
	{
		d_file_line_widget->Increment_Active_Line();
	}
}

void TDownloadEditor::Scale_Resolution_Changed(
	int									resolution)
{
	QString								line_text;
	
	// update existing editor data if it has changed
	if((d_active_line_index < 0) == false && d_data_editor_widget->Data_Item_Changed() == true)
	{
		this->Update_Active_Line();
		d_data_changed = true;
	}
	
	d_data_editor_widget->Set_Scale_Resolution(resolution);
	
	if(!(d_active_line_index < 0))
	{
		line_text = d_file_line_widget->Line(d_active_line_index);
		this->Process_Line(line_text);
	}
}

void TDownloadEditor::Read_File(void)
{
	QString								file_name;
	
	file_name = QFileDialog::getOpenFileName(this,
											QStringLiteral("Read Sharpe Download File"),
											d_active_file_label->text(),
											QStringLiteral("Sharpe Download Files (*.oad *.txt)"));
	
	if(file_name.length())
	{
		d_active_line_index = -1;
		d_active_command_id = -1;
		
		if(d_file_line_widget->Read_File(file_name))
		{
			d_save_file_action->setEnabled(true);
		}
		else
		{
			QMessageBox::warning(this,
								 QStringLiteral("Sharpe32 Download Editor"),
								 QStringLiteral("Unable to read download file"),
								 QMessageBox::Ok,
								 QMessageBox::Ok);
			
		}
		
		d_active_file_label->setText(QDir::toNativeSeparators(file_name));
		d_file_name_label->setText(QStringLiteral("Active Download File:"));

		d_data_changed = false;
	}
}

void TDownloadEditor::Write_File(void)
{
	QString								file_name;


	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("Write Sharpe Download File"),
											 QDir::fromNativeSeparators(d_active_file_label->text()),
											 QStringLiteral("Sharpe Download Files (*.oad *.txt)"));
	
	// update existing editor data if it has changed
	if((d_active_line_index < 0) == false && d_data_editor_widget->Data_Item_Changed() == true)
	{
		this->Update_Active_Line();
		d_data_changed = true;
	}

	if(file_name.length())
	{
		if(!d_file_line_widget->Write_File(file_name))
		{
			QMessageBox::warning(this,
								 QStringLiteral("Sharpe32 Download Editor"),
								 QStringLiteral("Error returned while trying to write a new download file"),
								 QMessageBox::Ok,
								 QMessageBox::Ok);

		}
		else
		{
			d_data_changed = false;
			d_active_file_label->setText(QDir::toNativeSeparators(file_name));
			d_file_name_label->setText(QStringLiteral("Active Download File:"));
		}
	}
}

void TDownloadEditor::Process_Line(
	const QString						&text)
{
	std::vector<TSH32DataEntryItem>::iterator entry_item_iter;
	TSH32DataEntry						sh32_data_entry;
	QRegExp								command_expression(QStringLiteral("([S][0-9ABCDEF]{1,2})"));	// S00 to SFF
	QString								line_text;
	QString								command_text;
	int									command_id;
	int									command_index;
	int									value;
	int									cntr;
	bool								valid;
	
	line_text = text;
	d_active_command_id = -1;
	
	command_index = command_expression.indexIn(line_text);
	
	if(!(command_index < 0) && command_expression.captureCount() == 1)
	{
		command_text = command_expression.cap(1);
		
		line_text.remove(0,command_index + command_text.length());
		
		command_id = this->Get_Command_Id(command_text);
		
		sh32_data_entry = d_sh32_data.Get_Entry(command_id);
		
		if(!(sh32_data_entry.command_id < 0))
		{
			d_active_command_id = sh32_data_entry.command_id;

			d_data_editor_widget->Reset(false);
			d_data_editor_widget->Set_Simple_Editor_Data(text);

			d_data_editor_widget->Set_Item_Widget_Count(sh32_data_entry.entry_items.size());
			
			for(entry_item_iter = sh32_data_entry.entry_items.begin(),cntr = 0;entry_item_iter != sh32_data_entry.entry_items.end();++entry_item_iter,++cntr)
			{
				value = this->Get_Entry_Value((*entry_item_iter).entry_id,line_text,&valid);
				d_data_editor_widget->Set_Item_Widget_Data(cntr,(*entry_item_iter).entry_id,value,(*entry_item_iter).description,valid,(*entry_item_iter).can_be_value);
			}
		}
		else
		{
			d_data_editor_widget->Reset(true);
			d_data_editor_widget->Set_Simple_Editor_Data(text);
		}
		
		return;
	}
	
	// unrecognized command
	d_data_editor_widget->Reset(true);
	d_data_editor_widget->Set_Simple_Editor_Data(text);
}

int TDownloadEditor::Get_Command_Id(
	const QString						&command_text) const
{
	QString								text(command_text);
	int									command_id(-1);
	bool								status;
	
	if(text.length() < 2)
	{
		return -1;
	}

	if(text[0] != QChar('S'))
	{
		return -1;
	}
	
	text.remove(0,1);
	
	command_id = static_cast<int>(text.toUInt(&status,16));
	
	if(status)
	{
		return command_id;
	}
	
	return -1;
}

int TDownloadEditor::Get_Entry_Value(
	const QChar							entry_id,
	const QString						&line,
	bool								*valid) const
{
	QString								line_text(line);
	QString								entry_text;
	QRegExp								entry_expression(QString("([%1][0-9ABCDEF]{0,8})").arg(entry_id));	// 0 to FFFFFFFF
	int									entry_index;
	int									value(0);
	
	assert(valid);
	*valid = false;
	
	entry_index = entry_expression.indexIn(line_text);

	if(!(entry_index < 0) && entry_expression.captureCount() == 1)
	{
		entry_text = entry_expression.cap(1);

		line_text.remove(0,entry_index + entry_text.length());
		

		if(entry_text.length() > 0)
		{
			entry_text.remove(0,1);	// remove entry text
			
			if(entry_text.length())
			{
				value = static_cast<int>(entry_text.toUInt(valid,16));
			}
			else // no value is valid
			{
				value = 0;
				*valid = true;
			}
		}
	}
	
	return value;
}

void TDownloadEditor::Update_Active_Line(void)
{
	std::vector<TSH32DataEntryItem>::iterator entry_item_iter;
	TSH32DataEntry						sh32_data_entry;
	int									cntr;
	QString								value;
	QString								line;
	QString								new_line;
	
	line = d_file_line_widget->Line(d_active_line_index);
	
	if(d_data_editor_widget->Using_Simple_Editor())
	{
		// new line is as entered by user
		new_line = d_data_editor_widget->Simple_Editor_Data();
		d_file_line_widget->Set_Line(d_active_line_index,new_line);
	}
	else
	{
		new_line = line;
		
		sh32_data_entry = d_sh32_data.Get_Entry(d_active_command_id);
		
		if(!(sh32_data_entry.command_id < 0))
		{
			for(entry_item_iter = sh32_data_entry.entry_items.begin(),cntr = 0;entry_item_iter != sh32_data_entry.entry_items.end();++entry_item_iter,++cntr)
			{
				this->Update_Parameter((*entry_item_iter).entry_id,d_data_editor_widget->Item_Widget_Data(cntr),&new_line);
			}
		}
		
		d_file_line_widget->Set_Line(d_active_line_index,new_line);
	}
}

bool TDownloadEditor::Update_Parameter(
	const QChar							&entry_id,
	const QString						&data,
	QString								* const line) const
{
	QRegExp								entry_expression(QString("([%1][0-9ABCDEF]{0,8})").arg(entry_id));	// 0 to FFFFFFFF
	int									entry_index;
	QString								entry_text;

	entry_index = entry_expression.indexIn(*line);
	
	if(!(entry_index < 0) && entry_expression.captureCount() == 1)
	{
		entry_text = entry_expression.cap(1);
		
		line->remove(entry_index,entry_text.length());
		line->insert(entry_index,QString("%1%2").arg(entry_id).arg(data));
		
		return true;
		
	}
	
	return false;
}



