/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2019  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QToolButton>
#include <QSizePolicy>
#include <QVBoxLayout>
#include <QWidget>
#include <QPaintEvent>
#include <QPainter>
#include <QRect>
#include <QPoint>
#include <QFontMetrics>
#include <QIcon>
#include <QPen>
#include <QSpacerItem>
#include <QFile>
#include <QEvent>
#include <QMouseEvent>
#include <QByteArray>
#include <QRegExp>
#include <QPalette>
#include <assert.h>

#include "filelinewidget.h"

TFileLineWidget::TFileLineWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QHBoxLayout							*control_hlayout;
	QSizePolicy							status_label_size_policy(QSizePolicy::Expanding,QSizePolicy::Preferred);
	QSizePolicy							display_widget_size_policy(QSizePolicy::Expanding,QSizePolicy::Expanding);
	QVBoxLayout							*widget_layout;
	QSpacerItem							*control_right_hspacer;
	QLabel								*scale_resolution_label;

	status_label_size_policy.setHorizontalStretch(0);
	status_label_size_policy.setVerticalStretch(0);

	display_widget_size_policy.setHorizontalStretch(0);
	display_widget_size_policy.setVerticalStretch(0);

	widget_layout = new QVBoxLayout(this);
	widget_layout->setContentsMargins(0,0,0,0);

	control_hlayout = new QHBoxLayout();
	control_hlayout->setContentsMargins(0,0,0,0);
	control_hlayout->setSpacing(0);

	d_up_button = new QToolButton(this);
	d_up_button->setAutoRepeat(true);
	control_hlayout->addWidget(d_up_button);

	d_down_button = new QToolButton(this);
	d_down_button->setAutoRepeat(true);
	control_hlayout->addWidget(d_down_button);
	
	control_right_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	control_hlayout->addItem(control_right_hspacer);

	scale_resolution_label = new QLabel(this);
	scale_resolution_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	control_hlayout->addWidget(scale_resolution_label);

	d_scale_resolution_edit = new QLineEdit(this);
	d_scale_resolution_edit->setMaximumWidth(80);
	d_scale_resolution_edit->setAlignment(Qt::AlignCenter);
	control_hlayout->addWidget(d_scale_resolution_edit);

	widget_layout->addLayout(control_hlayout);

	d_display_widget = new QWidget(this);
	d_display_widget->setMinimumWidth(100);
	d_display_widget->setMinimumHeight(100);
	d_display_widget->setSizePolicy(display_widget_size_policy);
	widget_layout->addWidget(d_display_widget);
	
	d_up_button->setIconSize(QSize(32,20));
	d_up_button->setMaximumSize(QSize(34,22));
	d_up_button->setIcon(QIcon(":/gui/tb_up.png"));
	
	d_down_button->setIconSize(QSize(32,20));
	d_down_button->setMaximumSize(QSize(34,22));
	d_down_button->setIcon(QIcon(":/gui/tb_down.png"));
	
	scale_resolution_label->setText(QStringLiteral("Scale Resolution:"));
	d_scale_resolution_edit->setText(QStringLiteral("1000"));
	
	d_display_widget->installEventFilter(this);
		
	d_display_font.setStyleHint(QFont::Monospace);
	d_display_widget->setFont(d_display_font);

	d_display_line_height = 1;
	d_display_initialize = true;
	
	d_active_line_index = 0;
	d_up_button->setEnabled(false);
	d_down_button->setEnabled(false);
	
	d_file_changed = false;
	
	connect(d_up_button,SIGNAL(clicked(void)),this,SLOT(Decrement_Active_Line(void)));
	connect(d_down_button,SIGNAL(clicked(void)),this,SLOT(Increment_Active_Line(void)));
	connect(d_scale_resolution_edit,SIGNAL(textEdited(const QString&)),this,SLOT(Scale_Resolution_Edited(const QString&)));
}

TFileLineWidget::~TFileLineWidget(void)
{
}

QString TFileLineWidget::Line(
	const int							index) const
{
	assert(!(index < 0));
	assert(index < d_file_lines.size());
	
	return d_file_lines[index];
}

int	TFileLineWidget::Scale_Resolution(void) const
{
	return abs(d_scale_resolution_edit->text().toInt());
}

void TFileLineWidget::Set_Line(
	const int							index,
	const QString						&text)
{
	assert(!(index < 0));
	assert(index < d_file_lines.size());
	
	d_file_lines[index] = text;
	d_display_widget->update();
	
	d_file_changed = true;
}

void TFileLineWidget::Set_Scale_Resolution(
	const int							resolution)
{
	if(resolution > 0)
	{
		d_scale_resolution_edit->setText(QString("%1").arg(resolution));
	}
}

bool TFileLineWidget::Read_File(
	const QString						&file_name)
{
	QFile								file(file_name);
	QByteArray							line;
	QString								line_text;
	QString								line_text_last;
	int									cntr;
	int									scale_resolution;
	int									index;
	QRegExp								resolution_expression(QString("^[Nn]+([0-9]{2,6})$"));	// N10000
	
	d_file_lines.clear();
	d_active_line_index = 0;
	
	if(!file.open(QIODevice::ReadOnly | QIODevice::Text))
	{
		return false;
	}
	
	cntr = 0;
	while(file.atEnd() == false && cntr < 128)
	{
		line = file.readLine();
		
		line_text = QString::fromLatin1(line).simplified();
		
		if(line_text.length())
		{
			line_text_last = line_text;
		}
		
		d_file_lines.push_back(line_text);
		cntr++;
	}
	
	file.close();
	
	
	index = resolution_expression.indexIn(line_text_last);
	
	if(!(index < 0) && resolution_expression.captureCount() == 1)
	{
		scale_resolution = resolution_expression.cap(1).toInt();
		d_scale_resolution_edit->setText(QString("%1").arg(scale_resolution));
		
		emit Scale_Resolution_Changed(scale_resolution);
	}
	else
	{
		scale_resolution = d_scale_resolution_edit->text().toInt();
	}
	
	if(d_file_lines.size())
	{
		emit Line_Activated(0);	// set default view to file line of newly opened file
	}
	
	this->update();

	d_up_button->setEnabled(false);
	d_down_button->setEnabled(true);

	return true;
}

bool TFileLineWidget::Write_File(
	const QString						&file_name)
{
	QFile								file(file_name);
	QStringList::const_iterator			iter;
	
	if(!file.open(QIODevice::WriteOnly))
	{
		return false;
	}
	
	for(iter = d_file_lines.begin();iter != d_file_lines.end();++iter)
	{
		file.write((*iter).toLatin1());
		file.write("\r\n");		// always use windows line terminators
	}
	
	file.close();
	
	return true;
}

void TFileLineWidget::Increment_Active_Line(void)
{
	d_active_line_index++;

	if(d_active_line_index < d_file_lines.size())
	{
		d_up_button->setEnabled(true);
		emit Line_Activated(d_active_line_index);
	}
	else
	{
		d_down_button->setEnabled(false);
		d_active_line_index = d_file_lines.size() - 1;
	}
	
	d_display_widget->update();
}

void TFileLineWidget::Decrement_Active_Line(void)
{
	d_active_line_index--;

	if(d_active_line_index < 0)
	{
		d_up_button->setEnabled(false);
		d_active_line_index = 0;
	}
	else
	{
		d_down_button->setEnabled(true);
		emit Line_Activated(d_active_line_index);
	}
	
	d_display_widget->update();
}

void TFileLineWidget::Set_Active_Line(
	int									line)
{
	if(line < 0)
	{
		return;
	}
	
	if(!(line < d_file_lines.size()))
	{
		return;
	}
	
	d_up_button->setEnabled(line != 0);
	d_down_button->setEnabled(line != (d_file_lines.size() - 1));
	
	d_active_line_index = line;
	emit Line_Activated(d_active_line_index);
	
	d_display_widget->update();
}

void TFileLineWidget::Scale_Resolution_Edited(
	const QString						&text)
{
	int									value;
	
	value = text.toInt();
	
	if(value > 0)
	{
		emit Scale_Resolution_Changed(value);
	}
}

bool TFileLineWidget::eventFilter(
	QObject								*object,
	QEvent								*event)
{
	QWidget								*widget(0);
	QMouseEvent							*mouse_press_event;
	
	if (event->type() == QEvent::Paint)
	{
		widget = qobject_cast<QWidget*>(object);
		
		if(widget)
		{
			object->event(event);	// draw the widget background
			this->Draw_Display_Widget(widget);
		}
		
		return true;
	}
	else if (event->type() == QEvent::MouseButtonPress)
	{
		widget = qobject_cast<QWidget*>(object);
		mouse_press_event = reinterpret_cast<QMouseEvent*>(event);
		
		if(widget && mouse_press_event)
		{
			this->Mouse_Click(widget,mouse_press_event->pos());
		}
		
		return true;
	}

	return QWidget::eventFilter(object, event);
}

void TFileLineWidget::Mouse_Click(
	QWidget								* const widget,
	const QPoint						&click_position)
{
	QRect								widget_rect;
	int									active_line_index;
	int									click_line_offset;
	int									line_count;
	int									data_index_start,data_index_end;
	
	if(d_display_initialize)
	{
		return;
	}

	widget_rect = widget->rect();

	line_count = widget_rect.height();
	line_count /= d_display_line_height;
	
	active_line_index = line_count / 2;
	
	data_index_start = d_active_line_index - line_count / 2;
	data_index_end = data_index_start + line_count;
	
	if(data_index_start < 0)
	{
		active_line_index += data_index_start;
		data_index_start = 0;
		data_index_end = line_count;
	}
	
	if(data_index_end > d_file_lines.size())
	{
		active_line_index += (data_index_end - d_file_lines.size());
		
		data_index_start -= (data_index_end - d_file_lines.size());
		if(data_index_start < 0)
		{
			active_line_index += data_index_start;
			data_index_start = 0;
		}
		
		data_index_end = d_file_lines.size();
	}
	
	click_line_offset = (click_position.y() - widget_rect.top()) / d_display_line_height;
	click_line_offset += data_index_start;
	
	this->Set_Active_Line(click_line_offset);
}

void TFileLineWidget::Draw_Display_Widget(
	QWidget								* const widget)
{
	QPainter							painter(widget);
	QPalette							palette;
	QRect								widget_rect;
	QRect								text_rect;
	QStringList::const_iterator			iter;
	int									draw_offset(0);
	int									active_line_index;
	int									active_line_offset;
	int									line_count;
	int									data_index_start,data_index_end;
	QPen								text_pen;//(QColor(0,0,0));
	QPen								text_highlight_pen;//(QColor(0,0,0));
	QPen								frame_pen;//(QColor(160,80,80));
	QPen								border_pen;//(QColor(196,196,196));
	QColor								highlight_color;
	
	palette = QApplication::palette();
	
	if(this->isEnabled())
	{
		text_pen.setColor(palette.color(QPalette::Active,QPalette::WindowText));
		text_highlight_pen.setColor(palette.color(QPalette::Active,QPalette::HighlightedText));
		frame_pen.setColor(palette.color(QPalette::Active,QPalette::Dark));
		border_pen.setColor(palette.color(QPalette::Active,QPalette::Shadow));
		highlight_color = palette.color(QPalette::Active,QPalette::Highlight);
	}
	else
	{
		text_pen.setColor(palette.color(QPalette::Disabled,QPalette::WindowText));
		text_highlight_pen.setColor(palette.color(QPalette::Disabled,QPalette::HighlightedText));
		frame_pen.setColor(palette.color(QPalette::Disabled,QPalette::Dark));
		border_pen.setColor(palette.color(QPalette::Disabled,QPalette::Shadow));
		highlight_color = palette.color(QPalette::Disabled,QPalette::Highlight);
	}
	
	painter.setFont(d_display_font);
	
	if(d_display_initialize)
	{
		QFontMetrics font_metrics = painter.fontMetrics();
		
		d_display_line_height = font_metrics.height();
		d_display_line_indent = font_metrics.averageCharWidth();
		
		d_display_initialize = false;
	}
	
	widget_rect = widget->rect();
	
	
	// nothing to draw if there is no data
	if(!d_file_lines.size())
	{
		return;
	}

	// draw outside frame
	painter.setPen(border_pen);
	painter.drawRect(widget_rect.left(),widget_rect.top(),widget_rect.width() - 1,widget_rect.height() - 1);

	line_count = widget_rect.height();
	line_count /= d_display_line_height;
	
	active_line_index = line_count / 2;
	
	data_index_start = d_active_line_index - line_count / 2;
	data_index_end = data_index_start + line_count;
	
	draw_offset = widget_rect.top();
	
	if(data_index_start < 0)
	{
		active_line_index += data_index_start;
		data_index_start = 0;
		data_index_end = line_count;
	}
	
	if(data_index_end > d_file_lines.size())
	{
		active_line_index += (data_index_end - d_file_lines.size());
		
		data_index_start -= (data_index_end - d_file_lines.size());
		if(data_index_start < 0)
		{
			active_line_index += data_index_start;
			data_index_start = 0;
		}
		
		data_index_end = d_file_lines.size();
	}
	
	active_line_offset = widget_rect.top() + active_line_index * d_display_line_height;

	for(iter = d_file_lines.begin() + data_index_start;iter != d_file_lines.begin() + data_index_end;++iter)
	{
		if(draw_offset == active_line_offset)
		{
			painter.setPen(frame_pen);
			painter.fillRect(0,draw_offset,widget_rect.width()-1,d_display_line_height,highlight_color);
			painter.drawRect(0,draw_offset,widget_rect.width()-1,d_display_line_height);
		}
		
		text_rect.setRect(d_display_line_indent,draw_offset,widget_rect.width(),d_display_line_height);
		
		if(draw_offset == active_line_offset)
		{
			painter.setPen(text_highlight_pen);
		}
		else
		{
			painter.setPen(text_pen);
		}
		painter.drawText(text_rect,Qt::AlignVCenter,(*iter));
		
		draw_offset += d_display_line_height;
	}
	
}
